package tests;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.concurrent.*;
import java.util.Scanner;

import cse332.exceptions.NotYetImplementedException;
import tests.exceptions.InformativeException;

/**
 * Contains all of the routines to conduct unit tests for a single file
 * @author anied, blank
 * @version 15su
 */
public abstract class TestsUtility {
    private ExecutorService executor;
    private Format fmt = new Format();
    private int all_points_earned = 0;
    private int all_points = 0;
    private StringBuffer print_output = new StringBuffer(); // Will append individual test results to this
    private StringBuffer informative_errors = new StringBuffer();
    private boolean hasException = false;
    private boolean hasTimeout = false;

    private static ByteArrayOutputStream testing_stream, main_stream;
     private static PrintStream Testing_out, Main_out;

    protected int POINTS = -1;
    protected int ALLOWED_TIME = 3000;

    private final String CLASSNAME = this.getClass().getName();
    private final String CLASSNAME_PRETTY = CLASSNAME.replaceAll("(PhaseA|PhaseB|Grading|Tests)", "").replaceAll("^.*\\.", "");

    protected boolean SHOW_FINAL_POINTS = CLASSNAME.startsWith("tests.grading.");
    protected boolean SHOW_TESTS = CLASSNAME.startsWith("tests.grading.");
    protected boolean SHOW_TEST_POINTS = false;
    protected boolean PRINT_STDERR = true;
    protected boolean PRINT_TESTERR =  CLASSNAME.startsWith("tests.grading.");
    protected boolean END_WITH_EXIT =  false;//!CLASSNAME.startsWith("tests.grading.");
    protected boolean DEBUG = false;
    public enum RESULT {
        PASSED, FAILED, EXCEPTION, TIMEOUT, MEMORY, SKIPPED, UNKNOWN;
        
        @Override
        public String toString() {
            switch(this) {
                case PASSED: return "✓ Passed";
                case FAILED: return "✘ Failed";
                case EXCEPTION: return "✝ Exception";
                case TIMEOUT: return "∞ Timeout";
                case MEMORY: return "ॐ OOMemory";
                case SKIPPED: return "";
                case UNKNOWN: return "‽ Unknown";
                default: throw new IllegalArgumentException();
            }
        }
    }

    private final int INT_ERROR_EXCEPTION = -10000;
    private final int INT_ERROR_MEMORY = -10001;
    
    protected abstract void run();
    
    /**
     * Get the cause of an exception
     * @param e The exception thrown
     * @return The original exception that caused the eventual exception
     */
    public static Throwable getCause(Throwable e) {
        Throwable cause = null; 
        Throwable result = e;

        while(null != (cause = result.getCause())  && (result != cause) ) {
            result = cause;
        }

        return result;
    }

    /**
     * Runs a unit test in a separate process, presuming it counts for 1 point
     * @param method The name of the method with the test
     * @return
     */
    protected int test(final String method) {
        return test(method, 1);
    }

    /**
     * Runs a unit test in a separate process
     * @param method The name of the method with the test
     * @param test_points The amount of points the test is
     * @return
     */
    protected int test(final String method, final int test_points) {
        testing_stream = new ByteArrayOutputStream();
        Testing_out = new PrintStream(testing_stream);
        
        int test_points_earned = 0;
        RESULT test_result = RESULT.UNKNOWN;

        try {
            testing_stream.flush();
        } catch (IOException e1) {
            e1.printStackTrace();
            // Do nothing
        }

        try {
            String clz = this.getClass().getName();

            if (executor == null) {
                executor = Executors.newCachedThreadPool();
            }


            String path = System.getProperty("java.class.path");

            final ProcessBuilder pb = new ProcessBuilder("java", "-classpath", path, "tests.TestRunner", clz, method);
            final Process p = pb.start();

            Callable<Object> task = new Callable<Object>() {
                @Override
                public Object call() throws InterruptedException {
                    int status = p.waitFor();

                    String output = "";
                    String error = "";
                
                    Scanner outputScanner = new Scanner(p.getInputStream()).useDelimiter("\\Z");
                    if (outputScanner.hasNext()) {
                        output = outputScanner.next();
                    }

                    Scanner errorScanner = new Scanner(p.getErrorStream()).useDelimiter("\\Z");
                    if (errorScanner.hasNext()) {
                        error = errorScanner.next();
                    }

                    if (PRINT_STDERR && error.length() > 0) {
                        addInformativeError(error);
                    }

                    if (output.startsWith("RESULT:")) {
                        return Integer.parseInt(output.substring("RESULT:".length()));
                    }
                    else if (output.startsWith("INFORMATIVE:")) {
                        addInformativeError(output.substring("INFORMATIVE:".length()));
                    }
                    else if (output.startsWith("UNEXPECTED:")) {
                        if (PRINT_TESTERR) {
                            String sep = System.lineSeparator();
                            addInformativeError("    Unexpected Exception Encountered: " +
                                CLASSNAME_PRETTY + "." + method + sep + "        " +
                                output.substring(output.indexOf(sep) + sep.length()).replaceAll(sep, sep + "    ")
                                );
                        }
                        
                        return INT_ERROR_EXCEPTION;
                    }
                    else if (DEBUG) {
                        // This means there was an invalid reponse from the child
                        // which is likely due to a timeout exception or the like.
                        System.err.println(output + ": " + status);
                    }

                    return 0;
                }
            };
            
            // Attempt the method, and catch timeouts
            Future<Object> future = executor.submit(task);
            try {
                test_points_earned = (int) future.get(ALLOWED_TIME, TimeUnit.MILLISECONDS); 
            } catch (TimeoutException e) {
                p.destroy();
                throw e;
            }
            
            // Gather the points earned
            if (test_points_earned <= INT_ERROR_EXCEPTION) {
                if (test_points_earned == INT_ERROR_MEMORY) {
                    test_result = RESULT.MEMORY; // Probably never going to happen given the current framework
                }
                else {
                    test_result = RESULT.EXCEPTION;
                }
                test_points_earned = 0;
            } 
            else if (test_points_earned == test_points) {
                test_result = RESULT.PASSED;
            }
            else {
                test_result = RESULT.FAILED;
            }
        } catch (IOException | IllegalArgumentException | ExecutionException | InterruptedException e) {
            System.err.println(e); // Our bad!
        } catch (TimeoutException e) {
            if (PRINT_TESTERR) {
                addInformativeError("Test Timed Out: " + CLASSNAME_PRETTY + "." + method + " had " + ALLOWED_TIME + "ms");
            }
            test_result = RESULT.TIMEOUT;
        }

        // Assign points
        switch (test_result) {
            case EXCEPTION:
                hasException = true;
            break;
            case TIMEOUT:
                hasTimeout = true;
            break;
            default:
                // pass
            break;
        }

        all_points_earned += test_points_earned;

        if (test_result != RESULT.SKIPPED) {
            all_points += test_points;
        }

        String resultStr = test_result.toString();
        if (SHOW_TEST_POINTS) {
            if (test_result == RESULT.PASSED || test_result == RESULT.FAILED) {
                resultStr = "";
            }
            resultStr += fmt.points(test_points_earned, test_points);
        }

        if (SHOW_TESTS) {
            print_output.append(fmt.justifiedLine("    " + method, resultStr));
        }

        if (informative_errors.length() > 0) {
            if (!print_output.toString().contains(informative_errors) || SHOW_TESTS) {
                print_output.append(fmt.justifiedLine((SHOW_TESTS ? "    " : "") + informative_errors));
            }
            informative_errors = new StringBuffer();
        }

        return test_points_earned; 
    }
    
    /**
     * Adds a message straight to the output
     * @param message
     */
    protected void addMessage(String message) {
        print_output.append(message);
    }

    /**
     * Add an informative message for the particular test
     * @param msg String of the message
     */
    protected void addInformativeError(String msg) {
        informative_errors.append(msg);
    }

    /**
     * Determines if the test suite was successful
     * Prints out the results
     * Potentially runs System.exit if it is supposed to
     */
    protected void finish() {
        boolean success = all_points_earned == all_points;
        print(success);
        if (END_WITH_EXIT) {
            System.exit(success ? 0 : 1);
        }
    }

    public static <T> T call(Object o, String methodName, Object... args) {
        Class[] argClasses = new Class[args.length];
        
        for (int i = 0; i < args.length; i++) {
            argClasses[i] = args[i].getClass();
        }

        try {
            Method method = o.getClass().getDeclaredMethod(methodName, argClasses);
            method.setAccessible(true);
         
            return (T)method.invoke(o, args);
        } catch (Exception e) {
            return null;
        }
    }

    private static void definalize(Field field) {
        try {
            Field modifiersField = Field.class.getDeclaredField("modifiers");
            modifiersField.setAccessible(true);
            modifiersField.setInt(field, field.getModifiers() & ~Modifier.FINAL);
        } catch (Exception e) {}
    }
    
    /**
     * Get a field from an object
     * @param o Object you want to get the field from
     * @param fieldName Name of the field
     * @return
     */
    @SuppressWarnings("unchecked")
    protected static <T> T getField(Object o, String fieldName) {
        try {
            Field field = o.getClass().getSuperclass().getDeclaredField(fieldName);
            field.setAccessible(true);
            definalize(field);
            Object f = field.get(o);
            return (T) f;
        } catch (Exception e) {
            try {
                Field field = o.getClass().getDeclaredField(fieldName);
                field.setAccessible(true);
                definalize(field);
                Object f = field.get(o);
                return (T) f;
            } catch (Exception e2) {
                return null;
            }
        }
    }
  
    /**
     * Set a field by name
     * @param o Object you want to get the field from
     * @param fieldName Name of the field
     * @return
     */
    @SuppressWarnings("unchecked")
    protected static void setField(Object o, String fieldName, Object value) {
        try {
            Field field = o.getClass().getSuperclass().getDeclaredField(fieldName);
            field.setAccessible(true);
            definalize(field);
            field.set(o, value);
        } catch (Exception e) {}
        try {
            Field field = o.getClass().getDeclaredField(fieldName);
            field.setAccessible(true);
            definalize(field);
            field.set(o, value);
        } catch (Exception e2) {}
        
    }

    /**
     * Set a field by type
     * @param o Object you want to get the field from
     * @param fieldName Name of the field
     * @return
     */
    @SuppressWarnings("unchecked")
    protected static void setField(Object o, Class type, Object value) {
        try {
            Field[] fields = o.getClass().getSuperclass().getDeclaredFields();
            for (Field f : fields) {
                if (type.isAssignableFrom(f.getType())) {
                    f.setAccessible(true);
                    definalize(f);
                    f.set(o, value);
                }
            }
        } catch (Exception e) {}
        try {
            Field[] fields = o.getClass().getDeclaredFields();
            for (Field f : fields) {
                if (type.isAssignableFrom(f.getType())) {
                    f.setAccessible(true);
                    definalize(f);
                    f.set(o, value);
                }
            }
        } catch (Exception e) {}
    }



    /**
     * Print the result of the test and any additional print_output
     * @param successes If success has happened
     */
    private void print(boolean success) {
        String success_str = "";
        success_str += hasException ? ", " + RESULT.EXCEPTION : "";
        success_str += hasTimeout ? ", " + RESULT.TIMEOUT : "";

        // Figure out the final points
        if (POINTS == 0) {
            all_points = 0;
        }
        else if (POINTS > 0) {
            all_points_earned = all_points_earned * POINTS / all_points;
            all_points = POINTS;
        }
        
        if (SHOW_FINAL_POINTS) {
            success_str += fmt.points(all_points_earned, all_points);
        } 
        else {
            success_str += ", ";
            success_str += success ? RESULT.PASSED : RESULT.FAILED;
            success_str += " Test(s)";
        }
        
        System.out.print(fmt.justifiedLine(CLASSNAME_PRETTY, success_str.substring(2)));
        if (print_output.length() > 0) {
            System.out.print(print_output);
        }
    }

    /**
     * Run the main method of a class, with no arguments
     * @param clazz The class you are calling
     * @return The console output made by the class main
     */
    public static String runMain(Class<?> clazz) {
        return runMain(clazz, null);
    }
    
    /**
     * Run the main method of a class
     * @param clazz The class you are calling
     * @param args Arguments the main method should take
     * @return The console output made by the class main
     */
    public static String runMain(Class<?> clazz, String[] args) {
        main_stream = new ByteArrayOutputStream();
        Main_out = new PrintStream(main_stream);
        
        String output = "";
        try {
            main_stream.flush();
            System.setOut(Main_out);
            clazz.getMethod("main",  String[].class).invoke(null, (Object) args);
            output = main_stream.toString();
            main_stream.reset();
        } catch (NotYetImplementedException | NoSuchMethodException | IOException | IllegalAccessException e) {
            // Do nothing
        } catch (InvocationTargetException e) {
            // Tell students why they failed!
            e.getCause().printStackTrace();
        }
        System.setOut(Testing_out);
        
        return output;
    }

    /**
     * Get an instance of a class given possible arguments. This will pick the first
     * constructor for the class and give it arguments matched to the types of arguments
     * you have given the constructor. This is useful if student writes a class that has
     * optional constructor arguments, like whether to hard code a comparator or get one
     * from the calling class. This helps prevent compilation errors.
     * @param clazz The class you want an instance of
     * @param possible_args An array (or variable number of inputs) of possible arguments
     * @return A new instance of the class
     */
    protected static <T> T getInstanceVargs(Class<?> clazz, Object... possible_args) {
        Constructor<?>[] constructors = clazz.getConstructors();
        
        // Grabs the new arguments
        Class<?>[] arg_classes = constructors[0].getParameterTypes();
        Object[] args = new Object[arg_classes.length];
        
        for (int i = 0; i < arg_classes.length; i++) {
            Class<?> arg_class = arg_classes[i];
            if (arg_class.toString().equals("int")) arg_class = Integer.class;
            if (arg_class.toString().equals("double")) arg_class = Double.class;
            if (arg_class.toString().equals("boolean")) arg_class = Boolean.class;
            if (arg_class.toString().equals("char")) arg_class = Character.class;
            
            for (Object arg : possible_args) {
                if (arg_class.isInstance(arg)) {
                    args[i] = arg;
                    break;
                }
            }
        }
        return getInstance(clazz, args);
    }
    
    /**
     * Get an instance of a class given arguments
     * @param clazz The class you want to get an instance of
     * @param args The arguments the class takes, as an array or just a variable number of arguments
     * @return An instance of the class
     */
    @SuppressWarnings("unchecked")
    protected static <T> T getInstance(Class<?> clazz, Object... args) {
        Constructor<?>[] constructors = clazz.getConstructors();
        
        // Try to make this constructor
        try {
            return (T) constructors[0].newInstance(args);
        } catch (InvocationTargetException | InstantiationException | IllegalAccessException e) {
            return null;
        } catch (IllegalArgumentException e) {
            throw new InformativeException("Constructor has incorrect arguments");
        }
    }
    

    /**
     * Get an instance of a class given arguments
     * @param clazz_str A name of the class you want to get an instance of
     * @param args The arguments the class takes, as an array or just a variable number of arguments
     * @return An instance of the class
     */
    @SuppressWarnings("unchecked")
    protected static <T> T getInstance(String clazz_str, Object... args) {
        try {
            return getInstance(Class.forName(clazz_str), args);
        } catch (ClassNotFoundException e) {
            throw new InformativeException("Class not found");
        }
    }
}
